static const char rcsid[] = "$Id: pop_sasl.c,v 1.3 2000/09/03 13:28:47 jurekb Exp $";
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "includes.h"

#include <saslutil.h>
#include <sasl.h>
#include <fcntl.h>
#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include <sys/resource.h>
#include <sys/stat.h>
#include <pwd.h>

struct secret_str {
	int len;
	char secret[SASLSECRETLN];
};

void print_errors(char *name, int result) {
	fprintf(stderr, "%s: %s\n", name, sasl_errstring(result, NULL, NULL));
}

int my_getsecret(void *context, const char *mechanism,
	const char *auth_identity, const char *realm, sasl_secret_t **secret) {
	return SASL_NOUSER;
}

int my_log(void *context, int priority, const char *message) {
	switch (priority) {
		case SASL_LOG_ERR:
			fprintf(stderr, "error: ");
			break;
		case SASL_LOG_WARNING:
			fprintf(stderr, "warning: ");
			break;
		default: ;
	}
	fprintf(stderr, "%s\n", message);
	return SASL_OK;
}

int my_putsecret(void *context, const char *mechanism,
	const char *auth_identity, const char *realm,
	const sasl_secret_t *secret) {
/* we ignore realm and auth_identity arguments */

	struct secret_str *my_secret = context;
	int len;

	if ((my_secret == NULL) || (secret == NULL))
		return SASL_FAIL;
	if ((secret->len + strlen(mechanism) +  sizeof(unsigned long int) +
		my_secret->len + 1) > sizeof(my_secret->secret))
		return SASL_BUFOVER;
	len = my_secret->len;
	strcpy(&my_secret->secret[len], mechanism);
	len += (strlen(mechanism) + 1);
	*((unsigned long int *)(&my_secret->secret[len])) = secret->len;
	len += sizeof(unsigned long int);
	memcpy(&my_secret->secret[len], secret->data, secret->len);
	my_secret->len = (len + secret->len);
	return SASL_OK;
}

int encode_secret(char *username, char *password, char *realm, 
	char *result, int *length) {
	struct secret_str secret;
	sasl_callback_t callbacks[] = {
		{SASL_CB_SERVER_PUTSECRET, my_putsecret, &secret},
		{SASL_CB_SERVER_GETSECRET, my_getsecret, NULL},
		{SASL_CB_LOG, my_log, NULL},
		{SASL_CB_LIST_END, NULL, NULL}
	};
	sasl_conn_t *connection;
	int ret;

	secret.len = 0;
	if ((ret = sasl_server_init(callbacks, "pop_sasl")) != SASL_OK) {
		print_errors("sasl_server_init", ret);
		return -1;
	}
	if ((ret = sasl_server_new("pop_sasl", NULL, realm, NULL,
		SASL_SECURITY_LAYER, &connection)) != SASL_OK) {
		print_errors("sasl_server_new", ret);
		sasl_done();
		return -1;
	}
	if ((ret = sasl_setpass(connection, username, password,
		strlen(password), 0, NULL)) != SASL_OK) {
		memset(&secret, 0, sizeof(secret));
		print_errors("sasl_setpass", ret);
		sasl_dispose(&connection);
		sasl_done();
		return -1;
	}
	sasl_dispose(&connection);
	sasl_done();
	if (secret.len > *length) {
		memset(&secret, 0, sizeof(secret));
		fprintf(stderr, "weird error: contact author of program\n");
		return -1;
	}
	memcpy(result, secret.secret, secret.len);
	*length = secret.len;
	memset(&secret, 0, sizeof(secret));
	return 0;
}

int main(int argc, char **argv) {
	char *pass;
	char username_buf[MAXARGLN + 1];
	char *username = NULL;
	char *realm = NULL;
	char passcopy[MAXARGLN + 1];
	char txtbuff[256];
	char result[SASLSECRETLN];
	char cfgfile[PATH_MAX];
	int length = sizeof(result);
	int fret, fd;
	struct stat stbuf;
	struct passwd *pwentry;
	struct rlimit corelimit = {0, 0};

	if (setrlimit(RLIMIT_CORE, &corelimit) < 0) {
		perror("setrlimit");
		exit(1);
	}
	if ((pwentry = getpwuid(getuid())) == NULL) {
		fprintf(stderr, "can't find user with UID: %u\n", getuid());
		exit(1);
	}
	while ((fret = getopt(argc, argv, "u:r:")) != EOF)
		switch(fret) {
			case 'u':
				username = optarg;
				break;
			case 'r':
				realm = optarg;
				break;
			case '?':
				exit(1);
			default:
				fprintf(stderr, "unexpected error - contact "
					"author of program\n");
				exit(1);
		}
	if (username == NULL) {
		username_buf[0] = 0;
		strncat(username_buf, pwentry->pw_name,
			sizeof(username_buf) - 1);
		if (strlen(pwentry->pw_name) >= sizeof(username_buf))
			fprintf(stderr, "Warning: username truncated\n");
		username = username_buf;
	}
	snprintf(txtbuff, sizeof(txtbuff),
		"Enter NEW SASL secret for user %.40s: ", username);
	pass = getpass(txtbuff);
	passcopy[0] = 0;
	strncat(passcopy, pass, sizeof(passcopy) - 1);
	memset(pass, 0, strlen(pass));
	snprintf(txtbuff, sizeof(txtbuff),
		"Reenter NEW SASL secret for user %.40s: ", username);
	pass = getpass(txtbuff);
	fret = strcmp(pass, passcopy);
	memset(pass, 0, strlen(pass));
	if (fret != 0) {
		memset(passcopy, 0, strlen(passcopy));
		fprintf(stderr, "SASL secrets don't match\n");
		exit(1);
	}
	if (stat(pwentry->pw_dir, &stbuf) < 0) {
		memset(passcopy, 0, strlen(passcopy));
		fprintf(stderr, "can't stat user home directory: %.1024s\n",
			pwentry->pw_dir);
		perror("stat");
		exit(1);
	}
	if (stbuf.st_mode & 2) {
		memset(passcopy, 0, strlen(passcopy));
		fprintf(stderr, "user home directory is world writable\n");
		exit(1);
	}
	if ((strlen(pwentry->pw_dir) + strlen(USERSASL) + 2) > sizeof(cfgfile)) {
		memset(passcopy, 0, strlen(passcopy));
		fprintf(stderr, "home directory name too long\n");
		exit(1);
	}
	strcpy(cfgfile, pwentry->pw_dir);
	strcat(cfgfile, "/");
	strcat(cfgfile, USERSASL);
	if ((fd = open(cfgfile, O_RDWR | O_CREAT | O_TRUNC, 0600)) < 0) {
		memset(passcopy, 0, strlen(passcopy));
		fprintf(stderr, "can't open user SASL file\n");
    		perror("open");
		exit(1);
	}
	if (fstat(fd, &stbuf) < 0) {
		memset(passcopy, 0, strlen(passcopy));
		close(fd);
		fprintf(stderr, "can't stat user SASL file\n");
		perror("fstat");
		exit(1);
	}
	if (!S_ISREG(stbuf.st_mode)) {
		memset(passcopy, 0, strlen(passcopy));
		close(fd);
		fprintf(stderr, "user SASL file is not regular file\n");
		exit(1);
	}
	if (stbuf.st_mode & 0177) {
		memset(passcopy, 0, strlen(passcopy));
		close(fd);
		fprintf(stderr, "user SASL file has wrong mode\n");
		exit(1);
	}
	if (encode_secret(username, passcopy, realm, result, &length) < 0) {
		memset(passcopy, 0, strlen(passcopy));
		close(fd);
		exit(1);
	}
	fret = write(fd, result, length);
	memset(passcopy, 0, strlen(passcopy));
	memset(result, 0, length);
	close(fd);
	if (fret < 0) {
		perror("write");
		exit(1);
	}
	exit(0);
}
